# !/usr/bin/perl
#
# p4_backup3.pl
#
# 機能概要：
# ■PERFORCE サーバルート・ディレクトリ配下のデータベースおよびディポを、
#   指定された場所にバックアップします。
# ■実行日が日曜日の場合、'p4d -jc -Z' により checkpoint を作成します。
#   実行日が日曜日以外の場合、'p4d -jj' により journal のローテートを行
#   います。
# ■実行日が日曜日のときはさらに、前週までの checkpoint と journal をす
#   べて削除します。
# ■-f オプションを指定した場合、曜日と関係なく（journal ではなく）チェ
#   ックポイントを作成します。ただし、前週までの checkpoint および 
#   journal は削除しません。
# ■バックアップの前には、データベースとディポ・ファイルとの整合性をチ
#   ェック（p4 verify）し、不整合が検出された場合はそれをログファイルに
#   記録します。このとき、バックアップの処理自体は続行します。
# ■実行の途中で "//depot/... - file(s) already have digests." というメ
#   ッセージが表示される場合がありますが、処理上の都合で表示されるもの
#   ですので、無視して問題ありません。
# ■環境設定は、本スクリプトと同じディレクトリにある p4_backup3.cfg フ
#   ィルから読み込みます。本スクリプトを実行する前には、必ず 
#   p4_backup3.cfg に対して適切な設定を行ってください。
# 　この p4_backup3.cfg には super 権限を持ったユーザのパスワードを記述
#   しなければなりませんので、このファイルを一般ユーザからは読み取れな
#   いような仕組みを、OS レベルで設定する必要があります。（例えば、一般
#   ユーザからは読み取り権限を剥奪するなど。）
# ■ログファイルは、バックアップ先ディレクトリの p4_backup3.log として
#   作成します。すでに存在していれば、既存のファイルに書き足します。
# ■以下の場合、本スクリプトはエラー終了します。
#   ・実行時パラメタが正しくない場合
# 　・ログファイルがオープンできなかった場合
# 　・環境設定ファイルがオープンできなかった場合
#   ・環境設定ファイルの設定が適切でなかった場合
# 　・バックアップ対象となる checkpoint または journal が作成できなかっ
#     た場合
#   ・バックアップのための rsync コマンドが異常終了した場合
# 　標準エラー出力を参照できない環境で本スクリプトを実行する場合、頻繁
# 　にログファイルの内容を確認してください。
#
# 使用方法：
# 　perl p4_backup3.pl [-f]
#
# 入力パラメタ
# 　なし
#
# オプション
#
#   -f  実行した曜日に関係なく、checkpoint を作成します。このとき、前週
#       の checkpoint や journal は削除しません。
#
# 終了コード：
# 　正常終了　0
# 　異常終了　1
#
# 注意事項：
# 本スクリプトの動作確認は、perl 5.10.0 で実施しました。
# 本スクリプトは、東陽テクニカが正式に販売し、サポート対象としている製
# 品ではありません。ご利用になる際は、この点をご了承いただき、十分に動
# 作検証を行った上でご利用ください。
#
# $Date: 2018/06/26 $	$Author: okano $
# $Id: //depot/perforce/p4_tools/p4_backup/Linux/p4_backup3.pl#6 $
# (c) Copyright 2009-2015 Toyo Corporation
#


###### 初期設定 ######
$ok = 0;
$ng = 1;
my $notexist = -1;
my $msg = 0;
my $fflag = 0;
$myself = `basename $0`;		# このファイル
chomp ($myself);
$bname = $myself; $bname =~ s/\.pl//;	# このファイルのベースの部分
$config_file = "$bname.cfg";		# 環境設定ファイル
$tmp_file1 = "$ENV{TEMP}/$bname" . "$$" . "_1.tmp";	# 一時ファイル
$info_head = "Info: $myself";		# 情報メッセージのヘッダ
$err_head = "Error: $myself";		# エラーメッセージのヘッダ
$p4title = "Perforce - The Fast Software Configuration Management System.";

# エラーメッセージ
$ErrMsg0 = "Please see Log File for further information.";
$ErrMsg1 = "Usage: jperl $myself [-f]";
$ErrMsg2 = "Cannot open log file.";
$ErrMsg3 = "Cannot backup Cehckpoint/Journal";
$ErrMsg4 = "Cannot create Cehckpoint/Journal.";
$ErrMsg5 = "xcopy command failed.";
$ErrMsg6 = "Invalid Perforce environment.";

#
# メイン処理
#
# パラメタのチェック
$argc = @ARGV;
if ($argc == 1) {
	if ($ARGV[0] eq "-f") {
		$fflag = 1;
	} else {
		die "$err_head: $ErrMsg1\n";
	}
} elsif ($argc != 0) {
	die "$err_head: $ErrMsg1\n";
}

# 構成ファイルから情報の取得
$rtn = &get_config();
if ($rtn != $ok) {
	exit $ng;
}
$port = $PORTNUM;			# PERFORCEポート（構成ファイルより）
$from_dir = $ROOTDIR;			# PERFORCEルート（構成ファイルより）
$to_dir = $TARGETDIR;			# バックアップ先（構成ファイルより）
$log_file = "$to_dir/$bname.log";	# ログファイルのパス

# logfileの準備
open(LOG, ">>$log_file") || die("$err_head: $ErrMsg2 - $log_file\n");

# p4 および p4d コマンドの構成（グローバル・オプションの設定）
if ($PASSWD eq "") {
	$P4 = "$_P4 -u $USER -p $port";
} else {
	$P4 = "$_P4 -u $USER -P $PASSWD -p $port";
}
$P4D = "$_P4D -r $from_dir";

# p4 および p4d コマンドのチェック
@p4com = 0;
@p4com = `$P4 -V`;
chomp (@p4com);
if ($p4com[0] ne $p4title) {
	&print_log(LOG, "$err_head: $ErrMsg6 - $P4\n");
	&settle;
	exit $ng;
}
@p4com = `$P4D -V`;
chomp (@p4com);
if ($p4com[0] ne $p4title) {
	&print_log(LOG, "$err_head: $ErrMsg6 - $P4D\n");
	&settle;
	exit $ng;
}

# 整合性のチェック（チェック）
print("p4 verify: Verifying...  ");
@vfy_list = `$P4 verify -q //depot/...`;
if ($#vfy_list != $notexist) {
	foreach $msg (@vfy_list) {
		&print_log(LOG, "$err_head: Verifying: $msg\n");
	}
}
print("Verified.\n");

# 整合性のチェック（記録）
@vfy_list = `$P4 verify -q -u //depot/...`;
print("p4 verify: Recording...  ");
print("Recorded.\n");

# 実行日の曜日（Sun, Mon, ...）を取得
$today = localtime(time);
chomp($today);
if ($today =~ /(\w*) .*/) {
	$day = $1;
}

# checkpoint および journal のパス
$ckpfile = $to_dir . "/$day";

# checkpoint の作成、または journal のローテート
print("p4d -jc|-jr: Creating...  ");
if ($day eq "Sun" || $fflag == 1) {
	# 日曜日は checkpoint の作成
	# -f オプション指定時も checkpoint の作成
	@ckp_list = `$P4D -jc -Z $ckpfile`;
	chomp(@ckp_list);
	if ($ckp_list[0] =~ /^Checkpointing to (.*)\.\.\./) {
		$ckp_with_num = $1;
	} else {
		foreach $msg (@ckp_list) {
			&print_log(LOG, "$err_head: $msg\n");
		}
		&print_log(LOG, "$err_head: $ErrMsg4\n");
		&settle;
		exit $ng;
	}
} else {
	# 日曜日以外は journal のローテート
	@ckp_list = `$P4D -jj $ckpfile`;
	chomp(@ckp_list);
	if ($ckp_list[0] =~ /^Rotating journal to (.*)\.\.\./) {
		$ckp_with_num = $1;
	} else {
		foreach $msg (@ckp_list) {
			&print_log(LOG, "$err_head: $msg\n");
		}
		&print_log(LOG, "$err_head: $ErrMsg4\n");
		&settle;
		exit $ng;
	}
}
print("Created.\n");
$ckp = `basename $ckp_with_num`;
chomp ($ckp);
&print_log(LOG, "$info_head: Back up $ckp from $from_dir to $to_dir.");

# 日曜日のときはさらに、前週の checkpoint および journal を削除
if ($day eq "Sun") {
	if (-f "$ckp_with_num") {
		print("Old Ckp/Jnl: Deleteing...  ");
		@del_ckp_list = `ls $to_dir/*.ckp.*`;
		chomp(@del_ckp_list);
		foreach $del_ckp (@del_ckp_list) {
			$del_ckp = `basename $del_ckp`;
			chomp ($del_ckp);
			if ($del_ckp ne $ckp) {
				system("rm -f \"$to_dir/$del_ckp\"");
			&print_log(LOG, "$info_head: Deleted \"$del_ckp\" from $to_dir.");
			}
		}
		system("rm -f $to_dir/*.jnl.*");
		print("Deleted.\n");
		&print_log(LOG, "$info_head: Deleted \"*.jnl.*\" from $to_dir.");
	} else {
		&print_log(LOG, "$err_head: $ErrMsg3\n");
		&settle;
		exit $ng;
	}
}
#P4ROOT直下のディレクトリをディポとみなす
@depot_list = `ls -F $from_dir | grep "/"`;
chomp(@depot_list);

# ディポ・ディレクトリをバックアップのためコピー
foreach $depot (@depot_list) {
	$depot =~ s/\///;
	print("Backup $depot: Copying...  ");
	$rtn = system("rsync -au \"$from_dir/$depot/\" \"$to_dir/$depot/\"");
	if ($rtn != $ok) {
		&print_log(LOG, "$err_head: $ErrMsg5 - rtn=$rtn\n");
		&settle;
		exit $ng;
	}
	&print_log(LOG, "$info_head: Back up depot \"$depot\" from $from_dir to $to_dir.");
}
close(LOG);
exit $ok;


#
# 構成ファイルから情報の取得
# 	IN  : なし
# 	OUT : 0 正常終了／1 異常終了
#
sub get_config {
	$my_dir = `dirname $0`;
	chomp ($my_dir);
	$cur_dir = `pwd`;
	chomp ($cur_dir);

	chdir("$my_dir");
	if (! -e $config_file) {
		return $ng;
	}
	require("$config_file");
	$dirbat = "$my_dir/$bname" . "_dir.bat";
	chdir("$cur_dir");
	return $ok;
}


#
# ログ情報を出力
# 	IN  : ファイルハンドル, 出力するメッセージ
# 	OUT : なし
#
sub print_log {
	my $FD = $_[0];
	my $trace_msg = "$_[1]";
	chomp($trace_msg);
	
	# 日時の取得
	($t_sec, $t_min, $t_hour, $t_mday, $t_mon, $t_year, 
			$tday, $t_ydat, $t_isdst) = localtime(time());
	
	# トレース情報を標準出力に出力
	printf($FD "*** %d/%02d/%02d %02d:%02d:%02d *** %s\n", 
			$t_year+1900, $t_mon+1, $t_mday, 
			$t_hour, $t_min, $t_sec, $trace_msg);
	return;
}


#
# 後始末処理
# 	IN  : なし
# 	OUT : なし
#
sub settle {
	close(LOG);
	print("$err_head: $ErrMsg0 - $log_file\n");
	return;
}
